#
# This file is autogenerated, do not edit manually. Instead,
# edit gen/BuffTrajPointStream_data.yml
#

import typing

from hal_impl.data import hal_data, NotifyDict

from .ctre_sim_enums import ErrorCode


class BuffTrajPointStream:

    def __init__(self):
        self.create_called = False

    def _create1(self) -> None:
        self.create_called = True

    def clear(self) -> ErrorCode:
        '''Clear all trajectory points.

        :returns: nonzero error code if operation fails.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        retval = ErrorCode.OK
        raise NotImplementedError("clear not implemented in simulation (file a bug on github if you need it)")

    def _write(self, position: float, velocity: float, arbFeedFwd: float, auxiliaryPos: float, auxiliaryVel: float, auxiliaryArbFeedFwd: float, profileSlotSelect0: int, profileSlotSelect1: int, isLastPoint: bool, zeroPos: bool, timeDur: int, useAuxPID: bool) -> ErrorCode:
        assert isinstance(position, (int, float))
        assert isinstance(velocity, (int, float))
        assert isinstance(arbFeedFwd, (int, float))
        assert isinstance(auxiliaryPos, (int, float))
        assert isinstance(auxiliaryVel, (int, float))
        assert isinstance(auxiliaryArbFeedFwd, (int, float))
        assert isinstance(profileSlotSelect0, int) and 0 <= profileSlotSelect0 < 1<<32
        assert isinstance(profileSlotSelect1, int) and 0 <= profileSlotSelect1 < 1<<32
        assert isinstance(isLastPoint, bool)
        assert isinstance(zeroPos, bool)
        assert isinstance(timeDur, int) and 0 <= timeDur < 1<<32
        assert isinstance(useAuxPID, bool)
        retval = ErrorCode.OK
        raise NotImplementedError("_write not implemented in simulation (file a bug on github if you need it)")





import tkinter as tk
from pyfrc.sim.ui_widgets import CheckButtonWrapper, Tooltip, ValueWidget

from hal_impl.data import hal_data

from ctre.basemotorcontroller import ControlMode


class CtreUI:

    can_mode_map = {v: k for k, v in ControlMode.__members__.items()}

    def update_tk_widgets(self, sim):

        for k, data in hal_data["CAN"].items():

            if data["type"] not in ["talonsrx", "victorspx"]:
                continue

            if not data["sim_display"]:
                self._add_CAN(sim, k, data)
                data["sim_display"] = True

            (motor, fl, rl, mode_lbl_txt, enc_txt, analog_txt, pwm_txt) = sim.can[k]
            data = hal_data["CAN"][k]
            mode = data["control_mode"]
            mode_lbl_txt.set(self.can_mode_map.get(mode, "Unknown"))
            # change how output works based on control mode
            if mode == ControlMode.PercentOutput:
                motor.set_value(data["value"])

            # elif mode == tsrxc.kMode_VoltCompen:
            #    #assume voltage is 12 divide by muliplier in cantalon code (256)
            #    motor.set_value(data['value']/12/256)

            # elif mode == tsrxc.kMode_SlaveFollower:
            # follow the value of the motor value is equal too
            #    try:
            #        followed = self.can[data['value']]
            #    except KeyError:
            #        value = 0
            #    else:
            #        value = followed[0].get_value()
            #
            #    motor.set_value(value)
            #
            # currently other control modes are not correctly implemented
            #
            else:
                motor.set_value(data["value"])

            enc_txt.set("E: %s" % round(data["quad_position"]))
            # analog_txt.set('A: %s' % round(data['analog_in_position']))
            # pwm_txt.set('P: %s' % data['pulse_width_position'])
            #
            ret = fl.sync_value(data["limit_switch_closed_for"])
            if ret is not None:
                data["limit_switch_closed_for"] = ret

            ret = rl.sync_value(data["limit_switch_closed_rev"])
            if ret is not None:
                data["limit_switch_closed_rev"] = ret

    def _add_CAN(self, sim, canId, device):

        # TODO: this is not flexible

        lcan = len(sim.can)
        row = (lcan % 6) * 2
        col = (lcan // 6) * 5

        lbl = tk.Label(sim.can_slot, text=str(canId))
        lbl.grid(column=col + 0, row=row)

        motor = ValueWidget(sim.can_slot, default=0.0)
        motor.grid(column=col + 1, row=row)
        sim.set_tooltip(motor, "CAN", canId)

        fl = CheckButtonWrapper(sim.can_slot, text="F")
        fl.grid(column=col + 2, row=row)

        rl = CheckButtonWrapper(sim.can_slot, text="R")
        rl.grid(column=col + 3, row=row)

        Tooltip.create(fl, "Forward limit switch")
        Tooltip.create(rl, "Reverse limit switch")

        mode_lbl_txt = tk.StringVar(value="")
        mode_label = tk.Label(sim.can_slot, textvariable=mode_lbl_txt)
        mode_label.grid(column=col + 4, row=row)

        labels = tk.Frame(sim.can_slot)
        labels.grid(column=col + 0, row=row + 1, columnspan=6)

        enc_value = tk.StringVar(value="E: 0")
        enc_label = tk.Label(labels, textvariable=enc_value)
        enc_label.pack(side=tk.LEFT)

        analog_value = tk.StringVar(value="A: 0")
        analog_label = tk.Label(labels, textvariable=analog_value)
        analog_label.pack(side=tk.LEFT)

        pwm_value = tk.StringVar(value="P: 0")
        pwm_label = tk.Label(labels, textvariable=pwm_value)
        pwm_label.pack(side=tk.LEFT)

        Tooltip.create(enc_label, "Encoder Input")
        Tooltip.create(analog_label, "Analog Input")
        Tooltip.create(pwm_label, "PWM Input")

        sim.can[canId] = (
            motor,
            fl,
            rl,
            mode_lbl_txt,
            enc_value,
            analog_value,
            pwm_value,
        )



# validated: 2018-01-14 DV a68a0c3ebf0b libraries/driver/include/ctre/phoenix/MotorControl/Faults.h
# -> ensure all the methods exist!


class FaultsBase:
    __slots__ = ("bits",)

    def __init__(self, bits: int = 0) -> None:
        self.bits = bits

    def toBitfield(self) -> int:
        return self.bits

    def hasAnyFault(self) -> bool:
        """True iff any of the flags are true."""
        return self.bits != 0

    def __str__(self):
        return " ".join(
            [
                "%s:%d" % (field, getattr(self, field))
                for field, method in self.__class__.__dict__.items()
                if isinstance(method, property)
            ]
        )



from ._impl import BuffTrajPointStream, ErrorCode

from .btrajectorypoint import BTrajectoryPoint

__all__ = ["BufferedTrajectoryPointStream"]


class BufferedTrajectoryPointStream(BuffTrajPointStream):
    """
        Stream of trajectory points for Talon/Victor motion profiling.
    """

    def __init__(self):
        super().__init__()
        self._create1()

    def write(self, trajPt: BTrajectoryPoint) -> ErrorCode:
        """Write a single trajectory point into the buffer.

        :param trajPt: Trajectory point to write.

        :returns: nonzero error code if operation fails.

        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.
        """
        return self._write(*trajPt)



import hal

from .autogen.faults import (
    CANifierFaults,
    CANifierStickyFaults,
    Faults,
    StickyFaults,
    PigeonIMU_Faults,
    PigeonIMU_StickyFaults,
)
from .motionprofilestatus import MotionProfileStatus

if hal.isSimulation():
    from .autogen.bufftrajpointstream_sim import BuffTrajPointStream
    from .autogen.canifier_sim import CANifier
    from .autogen.motcontroller_sim import MotController
    from .autogen.pigeonimu_sim import PigeonIMU
    from .autogen import ctre_sim_enums as _enum_module
else:
    from .ctre_roborio import BuffTrajPointStream, CANifier, MotController, PigeonIMU
    from . import ctre_roborio as _enum_module

# enums
ControlMode = _enum_module.ControlMode
DemandType = _enum_module.DemandType
ErrorCode = _enum_module.ErrorCode
NeutralMode = _enum_module.NeutralMode
RemoteFeedbackDevice = _enum_module.RemoteFeedbackDevice
RemoteSensorSource = _enum_module.RemoteSensorSource
FeedbackDevice = _enum_module.FeedbackDevice
FollowerType = _enum_module.FollowerType
InvertType = _enum_module.InvertType
SetValueMotionProfile = _enum_module.SetValueMotionProfile
StatusFrame = _enum_module.StatusFrame
StatusFrameEnhanced = _enum_module.StatusFrameEnhanced
VelocityMeasPeriod = _enum_module.VelocityMeasPeriod
RemoteLimitSwitchSource = _enum_module.RemoteLimitSwitchSource
LimitSwitchNormal = _enum_module.LimitSwitchNormal
LimitSwitchSource = _enum_module.LimitSwitchSource
ParamEnum = _enum_module.ParamEnum
CANifierStatusFrame = _enum_module.CANifierStatusFrame
CANifierControlFrame = _enum_module.CANifierControlFrame
GeneralPin = _enum_module.GeneralPin
PigeonIMU_StatusFrame = _enum_module.PigeonIMU_StatusFrame
PigeonIMU_ControlFrame = _enum_module.PigeonIMU_ControlFrame

del _enum_module



# validated: 2018-07-20 DV 3a78cd307217 java/src/com/ctre/phoenix/motorcontrol/can/BaseMotorController.java
# ----------------------------------------------------------------------------
#  Software License Agreement
#
#  Copyright (C) Cross The Road Electronics.  All rights
#  reserved.
#
#  Cross The Road Electronics (CTRE) licenses to you the right to
#  use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
#
#  THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
#  WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
#  LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
#  PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
#  CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL,
#  INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
#  PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
#  BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
#  THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
#  SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
#  (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
# ----------------------------------------------------------------------------

import typing

from .sensorcollection import SensorCollection
from .btrajectorypoint import BTrajectoryPoint
from .trajectorypoint import TrajectoryPoint
from ._impl import (
    ControlMode,
    DemandType,
    ErrorCode,
    Faults,
    FollowerType,
    InvertType,
    LimitSwitchNormal,
    MotController,
    MotionProfileStatus,
    NeutralMode,
    ParamEnum,
    RemoteFeedbackDevice,
    RemoteLimitSwitchSource,
    RemoteSensorSource,
    StatusFrame,
    StickyFaults,
    VelocityMeasPeriod,
)


__all__ = ["BaseMotorController"]


class BaseMotorController(MotController):
    """Base motor controller features for all CTRE CAN motor controllers."""

    ControlMode = ControlMode
    DemandType = DemandType
    InvertType = InvertType
    LimitSwitchNormal = LimitSwitchNormal
    NeutralMode = NeutralMode
    ParamEnum = ParamEnum
    RemoteFeedbackDevice = RemoteFeedbackDevice
    RemoteLimitSwitchSource = RemoteLimitSwitchSource
    RemoteSensorSource = RemoteSensorSource
    StatusFrame = StatusFrame
    VelocityMeasPeriod = VelocityMeasPeriod

    def __init__(self, arbId: int) -> None:
        """
        Constructor for motor controllers.
        
        :param arbId:
        """
        super().__init__()

        self._create1(arbId)
        self.arbId = arbId
        self.sensorColl = SensorCollection(self)
        self.controlMode = ControlMode.PercentOutput

    def getDeviceID(self) -> int:
        """
        Returns the Device ID
        
        :returns: Device number.
        """
        return self.getDeviceNumber()

    __set4_modes = frozenset(
        {
            ControlMode.PercentOutput,
            # ControlMode.TimedPercentOutput,
            ControlMode.Velocity,
            ControlMode.Position,
            ControlMode.MotionMagic,
            # ControlMode.MotionMagicArc,
            ControlMode.MotionProfile,
            ControlMode.MotionProfileArc,
        }
    )

    def set(
        self,
        mode: ControlMode,
        demand0: float,
        demand1Type: DemandType = DemandType.Neutral,
        demand1: float = 0.0,
    ):
        """
        Sets the appropriate output on the talon, depending on the mode.

        :param mode:
            The output mode to apply.
        :param demand0:
            The output value to apply. such as advanced feed forward and/or auxiliary close-looping in firmware.

            In :attr:`.ControlMode.PercentOutput`, the output is between -1.0 and 1.0, with 0.0 as
            stopped.
            
            In :attr:`.ControlMode.Voltage` mode, output value is in volts.
            
            In :attr:`.ControlMode.Current` mode, output value is in amperes.
            
            In :attr:`.ControlMode.Speed` mode, output value is in position change / 100ms.
            
            In :attr:`.ControlMode.Position` mode, output value is in encoder ticks or an analog value, depending on the sensor.
            
            In :attr:`.ControlMode.Follower` mode, the output value is the integer device ID of the talon to duplicate.
        :param demand1Type:
            The demand type for demand1.
            
            * Neutral: Ignore demand1 and apply no change to the demand0 output.
            * AuxPID: Use demand1 to set the target for the auxiliary PID 1.
            * ArbitraryFeedForward: Use demand1 as an arbitrary additive value to the
              demand0 output.  In PercentOutput the demand0 output is the motor output,
              and in closed-loop modes the demand0 output is the output of PID0.
        :param demand1:
            Supplemental value.  This will also be control mode specific for future features.
    
        Arcade Drive Example::
        
            _talonLeft.set(ControlMode.PercentOutput, joyForward, DemandType.ArbitraryFeedForward, +joyTurn)
            _talonRght.set(ControlMode.PercentOutput, joyForward, DemandType.ArbitraryFeedForward, -joyTurn)
                    
        Drive Straight Example::
                            
            # Note: Selected Sensor Configuration is necessary for both PID0 and PID1.
            _talonLeft.follow(_talonRght, FollowerType.AuxOutput1)
            _talonRght.set(ControlMode.PercentOutput, joyForward, DemandType.AuxPID, desiredRobotHeading)
                            
        Drive Straight to a Distance Example::
        
            # Note: Other configurations (sensor selection, PID gains, etc.) need to be set.
            _talonLeft.follow(_talonRght, FollowerType.AuxOutput1);
            _talonRght.set(ControlMode.MotionMagic, targetDistance, DemandType.AuxPID, desiredRobotHeading)
        
        """
        self.controlMode = mode

        if mode in self.__set4_modes:
            self._set_4(mode, demand0, demand1, demand1Type)
        elif mode is ControlMode.Follower:
            # did caller specify device ID
            if 0 <= demand0 <= 62:
                work = self.getBaseID()
                work >>= 16
                work <<= 8
                work |= int(demand0) & 0xFF
            else:
                work = int(demand0)
            self._set_4(mode, work, demand1, demand1Type)
        elif mode is ControlMode.Current:
            self.setDemand(mode, int(1000.0 * demand0), 0)  # milliamps
        else:
            self.setDemand(mode, 0, 0)

    def neutralOutput(self):
        """Neutral the motor output by setting control mode to disabled."""
        self.set(ControlMode.Disabled, 0, 0)

    def setInverted(self, invert: typing.Union[bool, InvertType]):
        """
        Inverts the hbridge output of the motor controller.

        This does not impact sensor phase and should not be used to correct sensor polarity.

        This will invert the hbridge output but NOT the LEDs.
        This ensures....

        - Green LEDs always represents positive request from robot-controller/closed-looping mode.
        - Green LEDs correlates to forward limit switch.
        - Green LEDs correlates to forward soft limit.
        
        :param invert:
            Invert state to set.
        """
        self.invert = int(invert)
        super()._setInverted_2(self.invert)

    def getInverted(self) -> bool:
        """:returns: invert setting of motor output"""
        if self.invert in (0, 1):
            return bool(self.invert)
        return super().getInverted()

    def getMotorOutputVoltage(self) -> float:
        """:returns: applied voltage to motor in volts"""
        return self.getBusVoltage() * self.getMotorOutputPercent()

    def getMotionProfileStatus(self) -> MotionProfileStatus:
        """
        Retrieve all status information.
        For best performance, Caller can snapshot all status information regarding the
        motion profile executer.
        """
        return MotionProfileStatus(*self._getMotionProfileStatus_2())

    def pushMotionProfileTrajectory(
        self, trajPt: typing.Union[TrajectoryPoint, BTrajectoryPoint]
    ) -> ErrorCode:
        """Push another trajectory point into the top level buffer (which is emptied
        into the motor controller's bottom buffer as room allows).

        :param trajPt: New point to push into buffer. You can use either the
                       :class:`.BTrajectoryPoint` or the legacy :class:`.TrajectoryPoint`
                       tuples. See the documentation for those structures for
                       details of the contents of each point.
        
        :returns: CTR_OKAY if trajectory point push ok. ErrorCode if buffer is
            full due to kMotionProfileTopBufferCapacity.
        
        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.
        """
        if len(trajPt) == 8:
            return self._pushMotionProfileTrajectory_2(*trajPt)
        else:
            return self._pushMotionProfileTrajectory_3(*trajPt)

    def getFaults(self) -> Faults:
        """Gets the last error generated by this object.

        Not all functions return an error code but can potentially report errors.
        This function can be used to retrieve those error codes.
        """
        return Faults(self._getFaults())

    def getStickyFaults(self) -> StickyFaults:
        """
        Polls the various sticky fault flags.
        """
        return StickyFaults(self._getStickyFaults())

    def getBaseID(self) -> int:
        return self.arbId

    def follow(
        self,
        masterToFollow: "BaseMotorController",
        followerType: FollowerType = FollowerType.PercentOutput,
    ):
        """
        Set the control mode and output value so that this motor controller will
        follow another motor controller. Currently supports following Victor SPX
        and Talon SRX.
        """
        id32 = masterToFollow.getBaseID()
        id24 = id32
        id24 >>= 16
        id24 = id24 & 0xFFFF
        id24 <<= 8
        id24 |= id32 & 0xFF

        if followerType == FollowerType.PercentOutput:
            self.set(ControlMode.Follower, id24)
        elif followerType == FollowerType.AuxOutput1:
            # follow the motor controller, but set the aux flag
            # to ensure we follow the processed output
            self.set(ControlMode.Follower, id24, DemandType.AuxPID, 0)
        else:
            self.neutralOutput()

    def valueUpdated(self):
        """
        When master makes a device, this routine is called to signal the update."""
        pass

    def getSensorCollection(self) -> SensorCollection:
        """
        :returns: object that can get/set individual raw sensor values."""
        return self.sensorColl

    def getControlMode(self) -> ControlMode:
        """
        :returns: control mode motor controller is in"""
        return self.controlMode

    def configAuxPIDPolarity(self, invert: bool, timeoutMs: int) -> ErrorCode:
        """Configures the Polarity of the Auxiliary PID (PID1).
        
        Standard Polarity:
        
        * Primary Output = PID0 + PID1
        * Auxiliary Output = PID0 - PID1
        
        Inverted Polarity:
        
        * Primary Output = PID0 - PID1
        * Auxiliary Output = PID0 + PID1
        
        :param invert:    If true, use inverted PID1 output polarity.
        
        :param timeoutMs: Timeout value in ms. If nonzero, function will wait for config
                          success and report an error if it times out. If zero, no
                          blocking or checking is performed.
        
        :returns: Error Code
        """
        return self.configSetParameter(
            ParamEnum.ePIDLoopPolarity, 1 if invert else 0, 0, 1, timeoutMs
        )



# validated: 2018-01-05 EN d376a57d7f94 java/src/com/ctre/phoenix/motorcontrol/can/VictorSPX.java
# ----------------------------------------------------------------------------
#  Software License Agreement
#
#  Copyright (C) Cross The Road Electronics.  All rights
#  reserved.
#
#  Cross The Road Electronics (CTRE) licenses to you the right to
#  use, publish, and distribute copies of CRF (Cross The Road) firmware files (*.crf) and Software
# API Libraries ONLY when in use with Cross The Road Electronics hardware products.
#
#  THE SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT
#  WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT
#  LIMITATION, ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A
#  PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL
#  CROSS THE ROAD ELECTRONICS BE LIABLE FOR ANY INCIDENTAL, SPECIAL,
#  INDIRECT OR CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF
#  PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
#  BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE
#  THEREOF), ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER
#  SIMILAR COSTS, WHETHER ASSERTED ON THE BASIS OF CONTRACT, TORT
#  (INCLUDING NEGLIGENCE), BREACH OF WARRANTY, OR OTHERWISE
# ----------------------------------------------------------------------------
import hal

from .basemotorcontroller import BaseMotorController

__all__ = ["VictorSPX"]


class VictorSPX(BaseMotorController):
    """VEX Victor SPX Motor Controller when used on CAN Bus.

    Consider using :class:`.WPI_VictorSPX`, which implements various WPILib interfaces.
    """

    def __init__(self, deviceNumber: int) -> None:
        """Constructor

        :param deviceNumber:
            [0,62]
        """
        super().__init__(deviceNumber | 0x01040000)
        hal.report(hal.UsageReporting.kResourceType_CTRE_future1, deviceNumber + 1)



#
# This file is autogenerated, do not edit manually. Instead,
# edit gen/PigeonIMU_data.yml
#

import typing

from hal_impl.data import hal_data, NotifyDict

from .ctre_sim_enums import ErrorCode


class PigeonIMU:

    def __init__(self):
        self.create_called = False

    def _create2(self, talonDeviceID: int) -> None:
        assert isinstance(talonDeviceID, int)
        self.create_called = True

    def _create1(self, deviceNumber: int) -> None:
        assert isinstance(deviceNumber, int)
        self.create_called = True

    def configSetParameter(self, param: int, value: float, subValue: int, ordinal: int, timeoutMs: int = 0) -> ErrorCode:
        '''Sets a parameter. Generally this is not used.

        This can be utilized in
        - Using new features without updating API installation.
        - Errata workarounds to circumvent API implementation.
        - Allows for rapid testing / unit testing of firmware.

        :param param:
            Parameter enumeration.
        :param value:
            Value of parameter.
        :param subValue:
            Subvalue for parameter. Maximum value of 255.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(param, int)
        assert isinstance(value, (int, float))
        assert isinstance(subValue, int) and 0 <= subValue < 1<<8
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("configSetParameter not implemented in simulation (file a bug on github if you need it)")

    def configGetParameter(self, param: int, ordinal: int, timeoutMs: int = 0) -> float:
        '''Gets a parameter. Generally this is not used.

        This can be utilized in

        - Using new features without updating API installation.
        - Errata workarounds to circumvent API implementation.
        - Allows for rapid testing / unit testing of firmware.

        :param param:
            Parameter enumeration.
        :param ordinal:
            Ordinal of parameter.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Value of parameter.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(param, int)
        assert isinstance(ordinal, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError("configGetParameter not implemented in simulation (file a bug on github if you need it)")

    def configGetParameter_6(self, param: int, valueToSend: int, ordinal: int, timeoutMs: int = 0) -> typing.Tuple[int, int]:
        assert isinstance(param, int) and -1<<31 <= param < 1<<31
        assert isinstance(valueToSend, int) and -1<<31 <= valueToSend < 1<<31
        assert isinstance(ordinal, int) and -1<<31 <= ordinal < 1<<31
        assert isinstance(timeoutMs, int) and -1<<31 <= timeoutMs < 1<<31
        raise NotImplementedError("configGetParameter_6 not implemented in simulation (file a bug on github if you need it)")

    def configSetCustomParam(self, newValue: int, paramIndex: int, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the value of a custom parameter. This is for arbitrary use.

        Sometimes it is necessary to save calibration/declination/offset
        information in the device. Particularly if the
        device is part of a subsystem that can be replaced.

        :param newValue:
            Value for custom parameter.
        :param paramIndex:
            Index of custom parameter [0-1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(newValue, int)
        assert isinstance(paramIndex, int)
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("configSetCustomParam not implemented in simulation (file a bug on github if you need it)")

    def configGetCustomParam(self, paramIndex: int, timoutMs: int) -> int:
        '''Gets the value of a custom parameter. This is for arbitrary use.

        Sometimes it is necessary to save calibration/declination/offset
        information in the device. Particularly if the
        device is part of a subsystem that can be replaced.

        :param paramIndex:
            Index of custom parameter [0-1].
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Value of the custom param.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(paramIndex, int)
        assert isinstance(timoutMs, int)
        raise NotImplementedError("configGetCustomParam not implemented in simulation (file a bug on github if you need it)")

    def configFactoryDefault(self, timeoutMs: int = 0) -> ErrorCode:
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("configFactoryDefault not implemented in simulation (file a bug on github if you need it)")

    def setYaw(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the Yaw register to the specified value.

        :param angleDeg: Degree of Yaw [+/- 23040 degrees]
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setYaw not implemented in simulation (file a bug on github if you need it)")

    def addYaw(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Atomically add to the Yaw register.

        :param angleDeg: Degrees to add to the Yaw register.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("addYaw not implemented in simulation (file a bug on github if you need it)")

    def setYawToCompass(self, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the Yaw register to match the current compass value.

        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setYawToCompass not implemented in simulation (file a bug on github if you need it)")

    def setFusedHeading(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the Fused Heading to the specified value.

        :param angleDeg: Degree of heading [+/- 23040 degrees]
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setFusedHeading not implemented in simulation (file a bug on github if you need it)")

    def addFusedHeading(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Atomically add to the Fused Heading register.

        :param angleDeg: Degrees to add to the Fused Heading register.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("addFusedHeading not implemented in simulation (file a bug on github if you need it)")

    def setFusedHeadingToCompass(self, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the Fused Heading register to match the current compass value.

        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setFusedHeadingToCompass not implemented in simulation (file a bug on github if you need it)")

    def setAccumZAngle(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the AccumZAngle.

        :param angleDeg: Degrees to set AccumZAngle to.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setAccumZAngle not implemented in simulation (file a bug on github if you need it)")

    def setTemperatureCompensationDisable(self, bTempCompDisable: int, timeoutMs: int = 0) -> ErrorCode:
        '''Disable/Enable Temp compensation. Pigeon has this on/False at boot.

        :param bTempCompDisable: Set to "False" to enable temperature compensation.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(bTempCompDisable, int)
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setTemperatureCompensationDisable not implemented in simulation (file a bug on github if you need it)")

    def setCompassDeclination(self, angleDegOffset: float, timeoutMs: int = 0) -> ErrorCode:
        '''Set the declination for compass. Declination is the difference between
        Earth Magnetic north, and the geographic "True North".

        :param angleDegOffset:  Degrees to set Compass Declination to.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDegOffset, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setCompassDeclination not implemented in simulation (file a bug on github if you need it)")

    def setCompassAngle(self, angleDeg: float, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the compass angle. Although compass is absolute [0,360) degrees, the
        continuous compass register holds the wrap-arounds.

        :param angleDeg:
            Degrees to set continuous compass angle to.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(angleDeg, (int, float))
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setCompassAngle not implemented in simulation (file a bug on github if you need it)")

    def enterCalibrationMode(self, calMode: int, timeoutMs: int = 0) -> ErrorCode:
        '''Enters the Calbration mode.  See the Pigeon IMU documentation for More
        information on Calibration.

        :param calMode: Calibration to execute
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for
            config success and report an error if it times out.
            If zero, no blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(calMode, int)
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("enterCalibrationMode not implemented in simulation (file a bug on github if you need it)")

    def _getGeneralStatus(self) -> typing.Tuple[int, int, int, int, float, int, int, int, int]:
        raise NotImplementedError("_getGeneralStatus not implemented in simulation (file a bug on github if you need it)")

    def getLastError(self) -> ErrorCode:
        '''Call GetLastError() generated by this object.

        Not all functions return an error code but can
        potentially report errors.

        This function can be used to retrieve those error codes.

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        retval = ErrorCode.OK
        raise NotImplementedError("getLastError not implemented in simulation (file a bug on github if you need it)")

    def get6dQuaternion(self) -> typing.List[float]:
        '''Get 6d Quaternion data.

        :param wxyz: [w, x, y, z]
        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("get6dQuaternion not implemented in simulation (file a bug on github if you need it)")

    def getYawPitchRoll(self) -> typing.List[float]:
        '''Get Yaw, Pitch, and Roll data.

        :param ypr_deg: [yaw, pitch, and roll]
        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getYawPitchRoll not implemented in simulation (file a bug on github if you need it)")

    def getAccumGyro(self) -> typing.List[float]:
        '''Get AccumGyro data.
        AccumGyro is the integrated gyro value on each axis.

        :param xyz_deg: [x, y, z]
        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getAccumGyro not implemented in simulation (file a bug on github if you need it)")

    def getAbsoluteCompassHeading(self) -> float:
        '''Get the absolute compass heading.

        :returns: compass heading [0,360) degrees.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getAbsoluteCompassHeading not implemented in simulation (file a bug on github if you need it)")

    def getCompassHeading(self) -> float:
        '''Get the continuous compass heading.

        :returns:
            continuous compass heading [-23040, 23040) degrees.
            Use SetCompassHeading to modify the wrap-around portion.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getCompassHeading not implemented in simulation (file a bug on github if you need it)")

    def getCompassFieldStrength(self) -> float:
        '''Gets the compass' measured magnetic field strength.

        :returns: field strength in Microteslas (uT).


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getCompassFieldStrength not implemented in simulation (file a bug on github if you need it)")

    def getTemp(self) -> float:
        '''Gets the temperature of the pigeon.

        :returns: Temperature in ('C)


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getTemp not implemented in simulation (file a bug on github if you need it)")

    def getState(self) -> int:
        '''Gets the current Pigeon state

        :returns: PigeonState enum


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getState not implemented in simulation (file a bug on github if you need it)")

    def getUpTime(self) -> int:
        '''Gets the current Pigeon uptime.

        :returns: How long has Pigeon been running in whole seconds. Value caps at 255.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getUpTime not implemented in simulation (file a bug on github if you need it)")

    def getRawMagnetometer(self) -> typing.List[int]:
        '''Get Raw Magnetometer data.

        :param rm_xyz: Array to fill with x[0], y[1], and z[2] data.
                       Number is equal to 0.6 microTeslas per unit.

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getRawMagnetometer not implemented in simulation (file a bug on github if you need it)")

    def getBiasedMagnetometer(self) -> typing.List[int]:
        '''Get Biased Magnetometer data.

        :param bm_xyz: Array to fill with x[0], y[1], and z[2] data.
                       Number is equal to 0.6 microTeslas per unit.

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getBiasedMagnetometer not implemented in simulation (file a bug on github if you need it)")

    def getBiasedAccelerometer(self) -> typing.List[int]:
        '''Get Biased Accelerometer data.

        :param ba_xyz: Array to fill with x[0], y[1], and z[2] data.
                       These are in fixed point notation Q2.14.  e.g. 16384 = 1G

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getBiasedAccelerometer not implemented in simulation (file a bug on github if you need it)")

    def getRawGyro(self) -> typing.List[float]:
        '''Get Raw Gyro data.

        :param xyz_dps: Array to fill with x[0], y[1], and z[2] data in degrees per second.

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getRawGyro not implemented in simulation (file a bug on github if you need it)")

    def getAccelerometerAngles(self) -> typing.List[float]:
        '''Get Accelerometer tilt angles.

        :param tiltAngles: Array to fill with x[0], y[1], and z[2] angles in degrees.

        :returns: The last ErrorCode generated.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getAccelerometerAngles not implemented in simulation (file a bug on github if you need it)")

    def _getFusedHeading2(self) -> typing.Tuple[int, int, float, int]:
        raise NotImplementedError("_getFusedHeading2 not implemented in simulation (file a bug on github if you need it)")

    def _getFusedHeading1(self) -> float:
        raise NotImplementedError("_getFusedHeading1 not implemented in simulation (file a bug on github if you need it)")

    def getResetCount(self) -> int:
        raise NotImplementedError("getResetCount not implemented in simulation (file a bug on github if you need it)")

    def getResetFlags(self) -> int:
        raise NotImplementedError("getResetFlags not implemented in simulation (file a bug on github if you need it)")

    def getFirmwareVersion(self) -> int:
        '''Gets the firmware version of the device.

        :returns:
            param holds the firmware version of the device. Device must be powered
            cycled at least once.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("getFirmwareVersion not implemented in simulation (file a bug on github if you need it)")

    def hasResetOccurred(self) -> bool:
        ''':returns: true iff a reset has occurred since last call.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        raise NotImplementedError("hasResetOccurred not implemented in simulation (file a bug on github if you need it)")

    def setLastError(self, value: int) -> ErrorCode:
        assert isinstance(value, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setLastError not implemented in simulation (file a bug on github if you need it)")

    def _getFaults(self) -> int:
        raise NotImplementedError("_getFaults not implemented in simulation (file a bug on github if you need it)")

    def _getStickyFaults(self) -> int:
        raise NotImplementedError("_getStickyFaults not implemented in simulation (file a bug on github if you need it)")

    def clearStickyFaults(self, timeoutMs: int = 0) -> ErrorCode:
        '''Clears the Sticky Faults

        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("clearStickyFaults not implemented in simulation (file a bug on github if you need it)")

    def setStatusFramePeriod(self, frame: int, periodMs: int, timeoutMs: int = 0) -> ErrorCode:
        '''Sets the period of the given status frame.

        :param statusFrame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int) and 0 <= periodMs < 1<<8
        assert isinstance(timeoutMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setStatusFramePeriod not implemented in simulation (file a bug on github if you need it)")

    def getStatusFramePeriod(self, frame: int, timeoutMs: int = 0) -> int:
        '''Gets the period of the given status frame.

        :param frame:
            Frame to get the period of.
        :param timeoutMs:
            Timeout value in ms. If nonzero, function will wait for config
            success and report an error if it times out. If zero, no
            blocking or checking is performed.
        :returns: Period of the given status frame.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(frame, int)
        assert isinstance(timeoutMs, int)
        raise NotImplementedError("getStatusFramePeriod not implemented in simulation (file a bug on github if you need it)")

    def setControlFramePeriod(self, frame: int, periodMs: int) -> ErrorCode:
        '''Sets the period of the given control frame.

        :param frame:
            Frame whose period is to be changed.
        :param periodMs:
            Period in ms for the given frame.
        :returns: Error Code generated by function. 0 indicates no error.


        .. note:: This function works on a real robot, but has not yet
                  been implemented in simulation mode. See :ref:`api_support`
                  for more details.

        '''
        assert isinstance(frame, int)
        assert isinstance(periodMs, int)
        retval = ErrorCode.OK
        raise NotImplementedError("setControlFramePeriod not implemented in simulation (file a bug on github if you need it)")





from ..faultsbase import FaultsBase


class CANifierFaults(FaultsBase):
    __slots__ = ()


class CANifierStickyFaults(FaultsBase):
    __slots__ = ()


class Faults(FaultsBase):
    __slots__ = ()

    @property
    def underVoltage(self) -> bool:
        return self.bits & (1 << 0) != 0

    @property
    def forwardLimitSwitch(self) -> bool:
        return self.bits & (1 << 1) != 0

    @property
    def reverseLimitSwitch(self) -> bool:
        return self.bits & (1 << 2) != 0

    @property
    def forwardSoftLimit(self) -> bool:
        return self.bits & (1 << 3) != 0

    @property
    def reverseSoftLimit(self) -> bool:
        return self.bits & (1 << 4) != 0

    @property
    def hardwareFailure(self) -> bool:
        return self.bits & (1 << 5) != 0

    @property
    def resetDuringEn(self) -> bool:
        return self.bits & (1 << 6) != 0

    @property
    def sensorOverflow(self) -> bool:
        return self.bits & (1 << 7) != 0

    @property
    def sensorOutOfPhase(self) -> bool:
        return self.bits & (1 << 8) != 0

    @property
    def hardwareESDReset(self) -> bool:
        return self.bits & (1 << 9) != 0

    @property
    def remoteLossOfSignal(self) -> bool:
        return self.bits & (1 << 10) != 0

    @property
    def apiError(self) -> bool:
        return self.bits & (1 << 11) != 0


class StickyFaults(FaultsBase):
    __slots__ = ()

    @property
    def underVoltage(self) -> bool:
        return self.bits & (1 << 0) != 0

    @property
    def forwardLimitSwitch(self) -> bool:
        return self.bits & (1 << 1) != 0

    @property
    def reverseLimitSwitch(self) -> bool:
        return self.bits & (1 << 2) != 0

    @property
    def forwardSoftLimit(self) -> bool:
        return self.bits & (1 << 3) != 0

    @property
    def reverseSoftLimit(self) -> bool:
        return self.bits & (1 << 4) != 0

    @property
    def resetDuringEn(self) -> bool:
        return self.bits & (1 << 5) != 0

    @property
    def sensorOverflow(self) -> bool:
        return self.bits & (1 << 6) != 0

    @property
    def sensorOutOfPhase(self) -> bool:
        return self.bits & (1 << 7) != 0

    @property
    def hardwareESDReset(self) -> bool:
        return self.bits & (1 << 8) != 0

    @property
    def remoteLossOfSignal(self) -> bool:
        return self.bits & (1 << 9) != 0

    @property
    def apiError(self) -> bool:
        return self.bits & (1 << 10) != 0


class PigeonIMU_Faults(FaultsBase):
    __slots__ = ()


class PigeonIMU_StickyFaults(FaultsBase):
    __slots__ = ()

